@@ -11,8 +11,8 @@ import threading
 import os
 import cryptography
 import grpc
-from pygnmi.spec.gnmi_pb2_grpc import gNMIStub
-from pygnmi.spec.gnmi_pb2 import (CapabilityRequest, Encoding, GetRequest,\
+from pygnmi.spec.v080.gnmi_pb2_grpc import gNMIStub
+from pygnmi.spec.v080.gnmi_pb2 import (CapabilityRequest, Encoding, GetRequest,\
     SetRequest, Update, TypedValue, SubscribeRequest, Poll, SubscriptionList,\
     SubscriptionMode, AliasList, UpdateResult)
 
@@ -41,7 +41,7 @@ class gNMIclient(object):
     def __init__(self, target: tuple, username: str = "", password: str = "",
                  debug: bool = False, insecure: bool = False, path_cert: str = None,
                  path_key: str = None, path_root: str = None, override: str = None,
-                 skip_verify = False, gnmi_timeout: int = 5, grpc_options: list = [],
+                 skip_verify = False, gnmi_timeout: int = 5, grpc_options: list = None,
                  show_diff: str = "", token: str = None, **kwargs):
 
         """
@@ -54,6 +54,8 @@ class gNMIclient(object):
         self.__path_cert = path_cert
         self.__path_key = path_key
         self.__path_root = path_root
+        if grpc_options is None:
+            grpc_options = []
         self.__options = ([('grpc.ssl_target_name_override', override)] + grpc_options) if override else grpc_options
         self.__token = token
         self.__gnmi_timeout = gnmi_timeout
@@ -117,29 +119,28 @@ class gNMIclient(object):
                     key = open(self.__path_key, 'rb').read()
                     root_cert = open(self.__path_root, 'rb').read()
 
-                except FileNotFoundError:
-                    logging.error('The SSL certificate cannot be opened.')
-                    raise Exception('The SSL certificate cannot be opened.')
+                except FileNotFoundError as e:
+                    logger.error('The SSL certificate cannot be opened.')
+                    raise gNMIException('The SSL certificate cannot be opened.', e)
 
             elif self.__path_cert:
                 try:
                     with open(self.__path_cert, 'rb') as f:
                         ssl_cert = f.read()
 
-                except FileNotFoundError:
+                except FileNotFoundError as e:
                     logger.error('The SSL certificate cannot be opened.')
-                    raise Exception('The SSL certificate cannot be opened.')
+                    raise gNMIException('The SSL certificate cannot be opened.', e)
 
             # Download a certficate from device if it is not provided
             else:
                 try:
                     ssl_cert = ssl.get_server_certificate((self.__target[0], self.__target[1])).encode("utf-8")
 
-                except:
+                except Exception as e:
                     logger.error(f'The SSL certificate cannot be retrieved from {self.__target}')
-                    raise Exception(f'The SSL certificate cannot be retrieved from {self.__target}')
+                    raise gNMIException(f'The SSL certificate cannot be retrieved from {self.__target}', e)
 
-            
             # Work with the certificate contents
             ssl_cert_deserialized = x509.load_pem_x509_certificate(ssl_cert, default_backend())
 
@@ -222,7 +223,7 @@ class gNMIclient(object):
             grpc.channel_ready_future(self.__channel).result(timeout=timeout)
 
         except grpc.FutureTimeoutError:
-            logging.error("Failed to setup gRPC channel, trying change cipher")
+            logger.error("Failed to setup gRPC channel, trying change cipher")
 
             try:
                 os.environ["GRPC_SSL_CIPHER_SUITES"] = "HIGH"
@@ -280,10 +281,8 @@ class gNMIclient(object):
             return response
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target_path}\nError: {err.details()}")
             logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
-
-            raise Exception(err)
+            raise gNMIException(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}", err)
 
         except:
             logger.error('Collection of Capabilities is failed.')
@@ -353,9 +352,9 @@ class gNMIclient(object):
         try:
             protobuf_prefix = gnmi_path_generator(prefix, target)
 
-        except:
+        except Exception as e:
             logger.error('Conversion of gNMI prefix to the Protobuf format failed')
-            raise Exception('Conversion of gNMI prefix to the Protobuf format failed')
+            raise gNMIException('Conversion of gNMI prefix to the Protobuf format failed', e)
 
         # Gnmi PATH
         try:
@@ -365,9 +364,9 @@ class gNMIclient(object):
             else:
                 protobuf_paths = [gnmi_path_generator(pe) for pe in path]
 
-        except:
+        except Exception as e:
             logger.error('Conversion of gNMI paths to the Protobuf format failed')
-            raise Exception ('Conversion of gNMI paths to the Protobuf format failed')
+            raise gNMIException('Conversion of gNMI paths to the Protobuf format failed', e)
 
         if self.__capabilities and 'supported_encodings' in self.__capabilities:
             if 'json' in self.__capabilities['supported_encodings']:
@@ -376,7 +375,7 @@ class gNMIclient(object):
                 pb_encoding = 4
 
         try:
-            gnmi_message_request = GetRequest(prefix=protobuf_prefix, path=protobuf_paths, 
+            gnmi_message_request = GetRequest(prefix=protobuf_prefix, path=protobuf_paths,
                                               type=pb_datatype, encoding=pb_encoding)
             debug_gnmi_msg(self.__debug, gnmi_message_request, "gNMI request")
 
@@ -457,13 +456,11 @@ class gNMIclient(object):
             return response
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target_path}\nError: {err.details()}")
             logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
+            raise gNMIException(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}", err)
 
-            raise Exception (err)
-
-        except:
-            logger.error(f'Collection of Get information failed is failed.')
+        # except:
+        #     logger.error(f'Collection of Get information failed is failed.')
 
             return None
 
@@ -500,29 +497,28 @@ class gNMIclient(object):
 
         if encoding not in encoding_set:
             logger.error(f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
-            raise Exception (f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
+            raise gNMIException(f'The encoding {encoding} is not supported. The allowed are: {", ".join(encoding_set)}.')
 
         # Gnmi PREFIX
         try:
             protobuf_prefix = gnmi_path_generator(prefix, target)
 
-        except:
+        except Exception as e:
             logger.error('Conversion of gNMI prefix to the Protobuf format failed')
-            raise Exception('Conversion of gNMI prefix to the Protobuf format failed')
+            raise gNMIException('Conversion of gNMI prefix to the Protobuf format failed', e)
 
         # Delete operation
         if delete:
             if isinstance(delete, list):
                 try:
                     del_protobuf_paths = [gnmi_path_generator(pe) for pe in delete]
-
-                except:
+                except Exception as e:
                     logger.error(f'Conversion of gNMI paths to the Protobuf format failed')
-                    raise Exception (f'Conversion of gNMI paths to the Protobuf format failed')
+                    raise gNMIException(f'Conversion of gNMI paths to the Protobuf format failed', e)
 
             else:
                 logger.error(f'The provided input for Set message (delete operation) is not list.')
-                raise Exception (f'The provided input for Set message (delete operation) is not list.')
+                raise gNMIException(f'The provided input for Set message (delete operation) is not list.')
 
         # Replace operation
         if replace:
@@ -545,11 +541,11 @@ class gNMIclient(object):
 
                     else:
                         logger.error(f'The input element for Update message must be tuple, got {ue}.')
-                        raise Exception (f'The input element for Update message must be tuple, got {ue}.')
+                        raise gNMIException(f'The input element for Update message must be tuple, got {ue}.')
 
             else:
                 logger.error(f'The provided input for Set message (replace operation) is not list.')
-                raise Exception ('The provided input for Set message (replace operation) is not list.')
+                raise gNMIException('The provided input for Set message (replace operation) is not list.')
 
         # Update operation
         if update:
@@ -572,11 +568,11 @@ class gNMIclient(object):
 
                     else:
                         logger.error(f'The input element for Update message must be tuple, got {ue}.')
-                        raise Exception (f'The input element for Update message must be tuple, got {ue}.')
+                        raise gNMIException(f'The input element for Update message must be tuple, got {ue}.')
 
             else:
                 logger.error(f'The provided input for Set message (update operation) is not list.')
-                raise Exception ('The provided input for Set message (replace operation) is not list.')
+                raise gNMIException('The provided input for Set message (replace operation) is not list.')
 
         try:
             # Adding collection of data for diff before the change
@@ -652,7 +648,6 @@ class gNMIclient(object):
                 return None
 
         except grpc._channel._InactiveRpcError as err:
-            print(f"Host: {self.__target_path}\nError: {err.details()}")
             logger.critical(f"GRPC ERROR Host: {self.__target_path}, Error: {err.details()}")
 
             return err
@@ -801,7 +796,7 @@ class gNMIclient(object):
         logger.info(f'Collecting Telemetry...')
 
         if (subscribe and poll) or (subscribe and aliases) or (poll and aliases):
-            raise Exception('Subscribe request supports only one request at a time.')
+            raise gNMIException('Subscribe request supports only one request at a time.')
 
         if poll:
             if isinstance(poll, bool):
@@ -854,7 +849,7 @@ class gNMIclient(object):
                 return self.subscribe_once(subscribe=subscribe, target=target)
 
         else:
-            raise Exception('Unknown subscription request mode.')
+            raise gNMIException('Unknown subscription request mode.')
 
     def subscribe_stream(self, subscribe: dict, target: str = None):
         if 'mode' not in subscribe:
@@ -1097,6 +1092,19 @@ class PollSubscriber(_Subscriber):
         return self._get_updates_till_sync(timeout=timeout)
 
 
+class gNMIException(Exception):
+    """Raised when a generic error in pygnmi occurred
+
+    Represents a generic pygnmi library error, described with a text what
+    the library tried to do. Optionally, an original exception can be added,
+    so the user of this library can further distinguish what happened in the
+    backend.
+    """
+    def __init__(self, message, orig_exc=None):
+        super().__init__(message)
+        self.orig_exc = orig_exc
+
+
 # User-defined functions
 def telemetryParser(in_message=None, debug: bool = False):
     """
